close all; clear all;
path(path, './blocks');

% *************************************************************
% parameters
% *************************************************************
subbands=[-2, 0, 3, 4, 5, 6, 7];
subbands=[0];
scSpacing=15e3;
subbandSpacing=333;


% *************************************************************
% PART I
% Link level model - operates on OFDM baseband stream
% *************************************************************

nFFT=32*512; % includes oversampling for distortion products etc, need at least 5*300
RBWwin=30e3; % spectrum analyzer resolution bandwidth (filter bandwidth)

nCp=floor(nFFT*0.07); % cyclic prefix length in samples
nSym=30; % CYCLIC signal stream (exactly as loop playback on a signal generator); one cycle holds this many symbols
scIndices=[]; % subcarrier indices of 4.5 / 5 MHz signal


% *************************************************************
% parameters
% *************************************************************

for s=subbands
  scIndices=[scIndices, s*subbandSpacing-150:s*subbandSpacing+149];
end

measBands=[-10:10];

if 0
  % for comparison, this is actual measurement data.
  calFactor1=9.2178819e-01  -9.0216521e-02i;
  calFactor3=7.2990867e-02   +8.3376073e-02i;
  calFactor5=-1.2484085e-02-1.5018770e-02i;
  calFactor7=3.9463450e-04+8.0113213e-04i;
end

if 1
  % pure 3rd order model, -33 dBc ACPR
  calFactor3=0.047;
end

% *************************************************************
% parameter-dependent constants
% *************************************************************
% IFFT indices for subcarriers relative to 0 Hz
IFFT_indices=mod(scIndices, nFFT)+1; % 0 maps to 1
rate=nFFT*scSpacing;
RBWpow=length(scIndices)*scSpacing;

% *************************************************************
% generate OFDM signal stream
% *************************************************************
stream=[];
for iSym=1:nSym
  % infinite-QAM :-) signal generator
  d = 1/sqrt(2)*transpose(randn(size(scIndices))+1i*randn(size(scIndices)));
  
  IFFT_input=zeros(1, nFFT);
  IFFT_input(IFFT_indices)=d;
  IFFT_output=ifft(IFFT_input)/sqrt(length(scIndices))*nFFT;
  
  stream=[stream, IFFT_output(end-nCp+1:end), IFFT_output]; % CP insertion
end

% *************************************************************
% Create raised-cosine window
% *************************************************************
winThis=linspace(0, 1, nCp+2);
winThis=winThis(2:end-1);

winThis=(1-cos(pi*winThis))/2;
winPrev=fliplr(winThis);
% *************************************************************
% apply raised-cosine windowing
% *************************************************************
symLen=nFFT+nCp;
for symIndex=1:nSym
  
  thisSymbolIndex=mod(symIndex-1, nSym)+1;
  prevSymbolIndex=mod(symIndex-2, nSym)+1;
  
  % sample indices pointing at the gap preceding "this" symbol
  thisGapFirst=(thisSymbolIndex-1)*symLen+1;
  thisGapLast=thisGapFirst+nCp-1;
  
  % sample indices used to extend "this" symbol periodically into the gap
  thisGapReplFirst=thisGapFirst+nFFT;
  thisGapReplLast=thisGapLast+nFFT;
  
  % sample indices pointing at the gap before the preceding symbol
  prevGapFirst=(prevSymbolIndex-1)*symLen+1;
  prevGapLast=prevGapFirst+nCp-1;
  
  % sample indices used to extend the preceding symbol periodically into the gap
  prevGapReplFirst=prevGapFirst+nCp;
  prevGapReplLast=prevGapLast+nCp;
  
  % retrieve periodic extensions 
  srcPrev=stream(prevGapReplFirst:prevGapReplLast);
  srcThis=stream(thisGapReplFirst:thisGapReplLast);

  % window and add
  xover=srcPrev .* winPrev + srcThis .* winThis;
  % replace current gap in stream
  stream(thisGapFirst:thisGapLast) = xover;
end
      
ref=stream; % this is the ideal (reference) signal
dist2=stream .* conj(stream); 
dist3=stream .* dist2;
%dist5=dist3 .* dist2;
%dist7=dist5 .* dist2;

% *************************************************************
% PA model, based on odd-order distortion.
% *************************************************************
PA_out = ref+calFactor3*dist3; %+calFactor5*dist5+calFactor7*dist7;

% error signal at output (remainder after minimum-least-squares projection on scaled/time-/phase shifted reference sig)
[dummyCoeff, proj]=mn_fitSignal_FFT(PA_out, ref);
err=PA_out-proj;

figure(3); grid on; hold on; xlabel('f/MHz'); ylabel('dB');
title('a bandlimited PA model using 3rd and 5th order terms');
[fr, fb]=mn_spectrum_local(ref, rate, RBWwin, RBWpow); 
plot(fb, fr, 'k', 'LineWidth', 3); 

[fr, fb]=mn_spectrum_local(PA_out, rate, RBWwin, RBWpow); 
plot(fb, fr, 'b', 'LineWidth', 2); 

[fr, fb]=mn_spectrum_local(err, rate, RBWwin, RBWpow); 
plot(fb, fr, 'r'); 

legend({'reference signal', 'PA output', 'distortion (uncorrelated with ref)'}, 'Location', 'NorthOutside');

if 0
  % *************************************************************
  % bandpass power measurement
  % *************************************************************
  e=fft(err);
  b=mn_fftFreqbase(length(err), rate);
  for m=measBands
    fMin=(m*subbandSpacing-150)*15e3;
    fMax=(m*subbandSpacing+150)*15e3;
    mask=zeros(size(b));
    mask(find(b > fMin))=1;
    mask(find(b > fMax))=0;  
    mask=fftshift(mask);
    
    e2=e .* mask;
    e2=ifft(e2);
    p=sum(e2 .* conj(e2)); % todo skip the ifft, simply scale
    p_dBc=10*log10(p/length(e2)*length(subbands)+1e-13);
    [m, p_dBc]
    plot([fMin/1e6, fMax/1e6], [p_dBc, p_dBc], '+-k', 'LineWidth', 2);
  end
end
  

% *************************************************************
% PART II
% "System" level model - operates on the power envelope alone
% *************************************************************

% sub-divide every subband into this many units
% use only odd numbers.
res=1;

% number of subbands in simulation bandwidth
r=30;
procBandsIndex=(-r*res:r*res)/res;
fd=zeros(size(procBandsIndex));

nBands=length(subbands)*res;
for s=subbands
  % find indices
  ix=find(abs(procBandsIndex-s)<0.45);
  % determine the actual number of bins, where power is allocated
  fd(ix)=1;
end 

% normalize
fd=fd/nBands;

% power amplifier model
% Refer to "spectral analysis of polynomial nonlinearity with
% applications to RF power amplifiers", G. Zhou and R. Raich, to 
% convert between "link level" and "system level" model parameters
% if needed
s1=1;
s3=4e-3;
% triple convolution via FFT
Q=fft(fd);
p3 = Q .* Q .* conj(Q); 
out=ifft(s1*Q+s3*p3);

for index=1:length(procBandsIndex)
  p=abs(out(index)); % "out" is positive and real-valued. Abs just in case.
  s=procBandsIndex(index);
  fMin=(s-0.5/res)*subbandSpacing*scSpacing;
  fMax=(s+0.5/res)*subbandSpacing*scSpacing;
  p_dBc=10*log10(p*nBands+1e-13);
  plot([fMin/1e6, fMax/1e6], [p_dBc, p_dBc], '+-g', 'LineWidth', 2);
end
